/*
 * [STA.java]
 *
 * Summary: Miscellaneous static methods for dealing with Strings in JDK 1.5+.
 *
 * Copyright: (c) 2007-2014 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.5 2005-07-14 split off from Misc, allow for compilation with old compiler.
 *  1.6 2005-10-15 add condense method.
 *  1.7 2006-03-04 format with IntelliJ and prepare Javadoc
 *  1.8 2006-03-05
 *  1.9 2008-03-10 add ST.firstWord
 *  2.0 2009-04-16 augment Common11 ST
 *  2.1 2009-04-30 add pruneExcessBlankLines
 *  2.2 2010-02-23 rename from ST to STA to avoid IntelliJ confusion.
 *  2.3 2011-11-10 add redorderLetters and dedupLetters
 */
package com.mindprod.common15;

import com.mindprod.common11.ST;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Arrays;

import static java.lang.System.out;

/**
 * Miscellaneous static methods for dealing with Strings in JDK 1.5+.
 * <p/>
 * Most are reimplementations of methods from com.mindprod.common1.ST using
 * the more efficient StringBuilder instead of StringBuffer.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 2.3 2011-11-10 add reorderLetters and dedupLetters
 * @since 2009-04-16
 */
public final class STA extends ST

    {
    // ------------------------------ CONSTANTS ------------------------------

    /**
     * true if you want extra debugging output and test code
     */
    private static final boolean DEBUGGING = false;

    // -------------------------- PUBLIC STATIC METHODS --------------------------

    /**
     * Collapse multiple whitespace in string down to a single space. Remove lead and trailing whitespace.
     * Earlier version collapsed only spaces, not whitespace
     *
     * @param s String to condense whitespace.
     *
     * @return String with all whitespace condensed and lead/trail whitespace removed.
     * @noinspection WeakerAccess, SameParameterValue
     * @see #squish(String)
     * @see com.mindprod.common11.ST#condense(String)
     */
    public static String condense( String s )
        {
        if ( s == null )
            {
            return null;
            }
        s = s.trim();

        final int len = s.length();
        if ( len == 0 )
            {
            return s;
            }
        // StringBuilder is faster than FastCat for char by char work
        StringBuilder b = new StringBuilder( len );
        boolean suppressSpaces = false;
        for ( int i = 0; i < len; i++ )
            {
            char c = s.charAt( i );
            if ( Character.isWhitespace( c ) )
                {
                if ( suppressSpaces )
                    {
                    // subsequent space
                    }
                else
                    {
                    // first space
                    b.append( ' ' );
                    suppressSpaces = true;
                    }
                }
            else
                {
                // was not a space
                b.append( c );
                suppressSpaces = false;
                }
            }// end for
        return b.toString();
        }

    /**
     * reorder the letters in a String in ascending order, case-sensitively, and then remove dups.
     * e.g. aCBBB" --> "BCa", "DAB" --> "ABD", " "Z X" --> "XZ", "   " --> ""
     *
     * @param field field to sort.
     *
     * @return field with letters composing it sorted in ascending order. Blanks squeezed out, but not dups.
     * @see #reorderLetters(String)
     */
    public static String deDupLetters( String field )
        {
        final String squished = STA.squish( field );
        if ( squished.length() <= 1 )
            {
            return field;
            }
        else
            {
            final char[] ca = squished.toCharArray();
            Arrays.sort( ca );  // case-sensitive sort letter by letter

            // most of the time, it will already be deDuped, check just in case.
            int dups = 0;
            for ( int i = 1; i < ca.length; i++ )
                {
                if ( ca[ i ] == ca[ i - 1 ] )
                    {
                    dups++;
                    }
                }

            if ( dups == 0 )
                {
                // no dups found, we are done.
                return String.valueOf( ca );
                }

            // we have to tediously dedup but at least we know the precise size of the final result ahead of time.
            final char[] deDuped = new char[ ca.length - dups ];

            if ( ca.length > 0 )
                {
                // always copy first char if there is one.
                deDuped[ 0 ] = ca[ 0 ];
                int j = 1;
                for ( int i = 1; i < ca.length; i++ )
                    {
                    if ( ca[ i ] != ca[ i - 1 ] )
                        {
                        deDuped[ j++ ] = ca[ i ];
                        }
                    }
                }
            return String.valueOf( deDuped );
            }
        }

    /**
     * Collapse multiple blank lines down to one.  Discards lead and trail blank lines.
     * Blank lines are lines that when trimmed have length 0.
     * Degraded version available in com.mindprod.common11.ST for JDK 1.1.
     *
     * @param lines               array of lines to tidy.
     * @param minBlankLinesToKeep usually 1 meaning 1+ consecutive blank lines become 1, effectively collapsing
     *                            runs of blank lines down to 1.
     *                            if 2, 1 blank line is removed, and 2+ consecutive blanks lines become 1,
     *                            effectively undouble spacing.
     *                            if zero, non-blank lines will be separated by one blank line, even if there was not
     *                            one there to begin with, completely independent of preexisting blank lines,
     *                            effectively double spacing..
     *                            9999 effectively removes all blank lines.
     *
     * @return array of lines with lead and trail blank lines removed, and excess blank lines collapsed down to one
     *         or 0.
     *         The results are NOT trimmed.
     */
    public static String[] pruneExcessBlankLines( String[] lines, int minBlankLinesToKeep )
        {
        int firstNonBlankLine = lines.length;
        for ( int i = 0; i < lines.length; i++ )
            {
            if ( lines[ i ].trim().length() > 0 )
                {
                firstNonBlankLine = i;
                break;
                }
            }
        int lastNonBlankLine = -1;
        for ( int i = lines.length - 1; i > 0; i-- )
            {
            if ( lines[ i ].trim().length() > 0 )
                {
                lastNonBlankLine = i;
                break;
                }
            }
        if ( firstNonBlankLine > lastNonBlankLine )
            {
            return new String[ 0 ];
            }
        // collapse blank lines in the middle chunk
        ArrayList<String> keep = new ArrayList<String>( lastNonBlankLine - firstNonBlankLine + 1 );
        int pendingBlankLines = 0;
        for ( int i = firstNonBlankLine; i <= lastNonBlankLine; i++ )
            {
            if ( lines[ i ].trim().length() == 0 )
                {
                pendingBlankLines++;
                }
            else
                {
                if ( pendingBlankLines >= minBlankLinesToKeep )
                    {
                    keep.add( "" );
                    }
                keep.add( lines[ i ] );  // we don't trim. That is up to caller.
                pendingBlankLines = 0;
                }
            }
        return keep.toArray( new String[ keep.size() ] );
        }

    /**
     * used to prepare SQL string literals by doubling each embedded ' and wrapping in ' at each end. Further quoting is
     * required to use the results in Java String literals. If you use PreparedStatement, then this method is not
     * needed. The ' quoting is automatically handled for you.
     *
     * @param sql Raw SQL string literal
     *
     * @return sql String literal enclosed in '
     * @noinspection WeakerAccess, SameParameterValue
     */
    public static String quoteSQL( String sql )
        {
        // StringBuilder is better than FastCat for char by char work.
        StringBuilder sb = new StringBuilder( sql.length() + 5 );
        sb.append( '\'' );
        for ( int i = 0; i < sql.length(); i++ )
            {
            char c = sql.charAt( i );
            if ( c == '\'' )
                {
                sb.append( "\'\'" );
                }
            else
                {
                sb.append( c );
                }
            }
        sb.append( '\'' );
        return sb.toString();
        }

    /**
     * reorder the letters in a String in ascending order, case-sensitively.
     * e.g. "DAB" --> "ABD", "aCBBB" --> "BBBCa", "Z X" --> "XZ", "   " --> ""
     *
     * @param field field to sort.
     *
     * @return field with letters composing it sorted in ascending order. Blanks squeezed out, but not dups.
     * @see #deDupLetters(String)
     */
    public static String reorderLetters( String field )
        {
        final String squished = STA.squish( field );
        if ( squished.length() <= 1 )
            {
            return field;
            }
        else
            {
            final char[] ca = squished.toCharArray();
            Arrays.sort( ca );  // case-sensitive sort letter by letter
            return String.valueOf( ca );
            }
        }

    /**
     * null-safe equals
     *
     * @param a first object/string, possibly null
     * @param b second object/string, possibly null.
     *
     * @return true if objects compare equal, null == null
     */
    public static boolean same( Object a, Object b )
        {
        if ( a == null )
            {
            return b == null;
            }
        else
            {
            return b != null && a.equals( b );
            }
        }

    /**
     * insert spaces between the letters
     *
     * @param s string of letters
     *
     * @return string with a space between each letter.
     */
    public static String spaceOut( String s )
        {
        final StringBuilder sb = new StringBuilder( s.length() * 2 - 1 );
        for ( int i = 0, n = s.length(); i < n; i++ )
            {
            if ( i != 0 )
                {
                sb.append( ' ' );
                }
            sb.append( s.charAt( i ) );
            }
        return sb.toString();
        }

    /**
     * Remove all spaces from a String. Does not touch other whitespace.
     *
     * @param s String to strip of blanks.
     *
     * @return String with all blanks, lead/trail/embedded removed.
     * @see #condense(String)
     */
    public static String squish( String s )
        {
        if ( s == null )
            {
            return null;
            }
        s = s.trim();
        if ( s.indexOf( ' ' ) < 0 )
            {
            return s;
            }
        int len = s.length();
        // StringBuilder is better than FastCat for char by char work.
        StringBuilder b = new StringBuilder( len - 1 );
        for ( int i = 0; i < len; i++ )
            {
            char c = s.charAt( i );
            if ( c != ' ' )
                {
                b.append( c );
                }
            }// end for
        return b.toString();
        }// end squish

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Dummy constructor
     * STA contains only static methods.
     */
    protected STA()
        {
        }

    // --------------------------- main() method ---------------------------

    /**
     * Test harness, used in debugging
     *
     * @param args not used
     */
    public static void main( String[] args )
        {
        if ( DEBUGGING )
            {
            out.println( condense( "   this  is   spaced.  " ) );
            out.println( trimLeading( "   trim   " ) );
            out.println( trimTrailing( "   trim   " ) );
            out.println( trimLeading( "*****t*r*i*m****", '*' ) );
            out.println( trimTrailing( "*****t*r*i*m****", '*' ) );
            out.println( toString( Color.RED ) );
            out.println( toHexString( -3 ) );
            out.println( toHexString( 3 ) );
            out.println( countLeading( "none", ' ' ) );
            out.println( countLeading( "*one***", '*' ) );
            out.println( countLeading( "\n\ntw\n\n\no\n\n\n\n", '\n' ) );
            out.println( countTrailing( "none", ' ' ) );
            out.println( countTrailing( "***one*", '*' ) );
            out.println( countTrailing( "\n\n\n\nt\n\n\n\nwo\n\n", '\n' ) );
            out.println( quoteSQL( "Judy's Place" ) );
            out.println( parseLongPennies( "$5.00" ) );
            out.println( parseLongPennies( "$50" ) );
            out.println( parseLongPennies( "50" ) );
            out.println( parseLongPennies( "$50-" ) );
            out.println( penniesToString( 0 ) );
            out.println( penniesToString( -1 ) );
            out.println( penniesToString( 20 ) );
            out.println( penniesToString( 302 ) );
            out.println( penniesToString( -100000 ) );
            out.println( toBookTitleCase( "handbook to HIGHER consciousness" ) );
            out.println( toBookTitleCase( "THE HISTORY OF THE U.S.A." ) );
            out.println( toBookTitleCase( "THE HISTORY OF THE USA" ) );
            out.println( rightPad( "abc", 6, true ) + "*" );
            out.println( rightPad( "abc", 2, true ) + "*" );
            out.println( rightPad( "abc", 2, false ) + "*" );
            out.println( rightPad( "abc", 3, true ) + "*" );
            out.println( rightPad( "abc", 3, false ) + "*" );
            out.println( rightPad( "abc", 0, true ) + "*" );
            out.println( rightPad( "abc", 20, true ) + "*" );
            out.println( rightPad( "abc", 29, true ) + "*" );
            out.println( rightPad( "abc", 30, true ) + "*" );
            out.println( rightPad( "abc", 31, true ) + "*" );
            out.println( rightPad( "abc", 40, true ) + "*" );
            out.println( toUpperCase( 'q' ) );
            out.println( toUpperCase( 'Q' ) );
            out.println( toUpperCase( "The quick brown fox was 10 feet tall." ) );
            out.println( toUpperCase( "THE QUICK BROWN FOX WAS 10 FEET TALL." ) );
            out.println( toUpperCase( "the quick brown fox was 10 feet tall." ) );
            out.println( toLowerCase( 'q' ) );
            out.println( toLowerCase( 'Q' ) );
            out.println( toLowerCase( "The quick brown fox was 10 feet tall." ) );
            out.println( toLowerCase( "THE QUICK BROWN FOX WAS 10 FEET TALL." ) );
            out.println( toLowerCase( "the quick brown fox was 10 feet tall." ) );
            out.println( "count instances should be 4: " + countInstances( " abab abcdefgab", "ab" ) );
            }
        }// end main
    }
